/*	Buzzer.cpp
*
*	Buzzer output using Timer/Counter 0 for frequency control.
*
*	NOTE: this code...
*	- uses Timer/Counter 0
*	- assumes internal clock is running at 8MHz
*
*	Written by Cathy Saxton
*	robotics@idleloop.com
*
*	Revisions:
*		6-29-06	signal.h/INTERRUPT() -> interrupt.h/ISR() [new avr-gcc]
*/

#include <avr/interrupt.h>
#include "main.h"

/*	want (max) 3.1kHz freq -- toggle at 6.2 kHz, every 165 us
	8 MHz uC clk, Timer/Counter 0 overflow at 256
	
	bits  clk/  freq     overflow every...
	----  ----  -------  ---------------
	001   /1     32 kHz   32 us
	010   /8      4 kHz  256 us
	011   /64   512  Hz    2 ms

	use clk/8 for overflow every 256 uS
	
	overflows  toggle every...  toggle rate  freq to buzzer
	---------  ---------------  -----------  --------------
	    1          256 us           4 kHz         2 kHz [higher pitch]
	    2          512 us           2 kHz         1 kHz
	    3          768 us         1.3 kHz       683  Hz [lower pitch]
*/

/* all this stuff is needed by the interrupt, which isn't a member
   of the class, thus the static declarations in this file */
static OUT *s_poutBuzzer;		// output port for the buzzer
static ulong s_msStart;			// start of buzz
static uint  s_msDuration;		// duration for buzz
static uchar s_cOverflow;		// number of overflows since last toggle
static uchar s_cOverflowToggle;	// number of overflows for each toggle

BUZZER::BUZZER(OUT *pout)
{
	if (TCCR0 != 0)
	{
		/* we were running another buzzer; stop it
		   before changing to this one */
		TCCR0 = 0;	// turn off timer
		ClearBit(TIMSK, TOIE0);
		s_poutBuzzer->SetLow();
	}
	s_poutBuzzer = pout;
	pout->SetLow();
}

/* sets up microcontroller for Timer/Counter 2; sets up globals for toggles */
void BUZZER::Buzz(uchar tone, uint ms) const
{
	/* note: keep in sync with tone enum in Buzzer.h */
	uchar mptonecOverflow[toneMax] = {1, 2, 3};
	
	if (TCCR0 != 0)
	{
		/* already running; shut it off so we don't get interrupted
		   while setting up for this one */
		TCCR0 = 0;	// turn off timer
		ClearBit(TIMSK, TOIE0);
		s_poutBuzzer->SetLow();
	}

	/* make sure tone is in valid range (won't be < min
	   since min = 0, and tone is unsigned) */
	if (tone >= toneMax)
		tone = toneMax;
	s_cOverflowToggle = mptonecOverflow[tone];
	s_cOverflow = 0;
	s_msStart = g_timer.MsCur();
	s_msDuration = ms;
	SetBit(TIMSK, TOIE0);
	TCNT0 = 0;
	TCCR0 = 0x02;	// clk/8, 4 kHz, overflow every 256 us
}

/* this interrupt happens when Timer/Counter 0 overflows */
ISR(TIMER0_OVF_vect)
{
	if (++s_cOverflow >= s_cOverflowToggle)
	{
		s_poutBuzzer->Toggle();
		s_cOverflow = 0;
	}

	if (g_timer.HasElapsed(s_msStart, s_msDuration))
	{
		TCCR0 = 0;	// turn off timer
		ClearBit(TIMSK, TOIE0);
		s_poutBuzzer->SetLow();
	}
}
