/*	LCD.cpp
*
*	Class for controlling the LCD.
*
*	Assumes the following data connections:
*	LCD		ATmega8
*	---		-------
*	DB7		PD7		[high bit]
*	DB6		PB5
*	DB5		PB4
*	DB4		PB3
*
*	Written by Cathy Saxton
*	robotics@idleloop.com
*/

#include <avr/io.h>
#include "main.h"

#define cchScreen 8	// number of chars (per line) visible on the screen

/* Register Select pin: 0 for cmd, 1 for data */
#define RSForCmd()		m_poutRS->SetLow()
#define RSForData()		m_poutRS->SetHigh()
/* Read/!Write pin: 0 for write, 1 for read */
#define RWForWrite()	m_poutRW->SetLow()
#define RWForRead()		m_poutRW->SetHigh()
/* E pin: enable on high transition */
#define ELow()			m_poutE->SetLow()
#define EHigh()			m_poutE->SetHigh()

/* output pins (D7, B5, B4, B3)
	output: 0000wxyz (uchar in WriteNibble)
	PORTD:  w-------
	PORTB:  --xyz---
	 (bits) 76543210
*/
#define mskPortB	0x38
#define mskPortD	0x80
#define shiftPortB	3
#define shiftPortD	4

/* LCD DB7 = Orangutan/ATmega8 PD7 (no pull-up R, so easy swap between in & out) */
#define DB7AsInput()	(DDRD &= ~0x80)
#define DB7AsOutput()	(DDRD |= 0x80)
#define DB7Val()		(PIND & 0x80)

/* LCD DB6/5/4 (PB5/4/3) also used for input buttons */
/* for LCD, need to output on ports connected to LCD's DB lines */
#define LCDDBAsOutput()	(DDRB | mskPortB)		// for LCD
/* for pushbuttons, need to disable pull-up resistors and set for input */
#define LCDDBAsInput()	(PORTB &= ~mskPortB, DDRB &= ~mskPortB)


bool LCD::s_fInit = false;

LCD::LCD(OUT *poutRS, OUT *poutRW, OUT *poutE)
	: m_poutRS(poutRS), m_poutRW(poutRW), m_poutE(poutE)
{
	DB7AsOutput();
	m_poutRS->SetLow();
	m_poutRW->SetLow();
	m_poutE->SetLow();

	if (!s_fInit)
		Init();
}

void LCD::Init() const
{
	/* make sure LCD has a chance to finish initializing after power-up */
	g_timer.WaitMs(20);

	/* send forced reset */
	WriteNibble(0, 0x03);	// (RS = 0, code = 0011)
	g_timer.WaitMs(5);
	WriteNibble(0, 0x03);	// (RS = 0, code = 0011)
	g_timer.WaitMs(1);
	WriteNibble(0, 0x03);	// (RS = 0, code = 0011)
	g_timer.WaitMs(1);

	WriteNibble(0, 0x02);	// (RS = 0, code = 0010)	// set to 4-bit operation
	g_timer.WaitMs(1);
	WriteCommand(0, 0x28);	// 4-bit operation, 2 lines (5x8)

	/* now we can use the status bit! */
	s_fInit = true;

	/* other initialization */
	WriteCommand(0, 0x06);	// increment, no shift
	ClearAndReturn();		// clear screen and return cursor home
	HideCursor();			// (this also turns on display)
}

/* prints given string on the specified line (1 or 2);
   remainder of (visible) line will be padded with spaces by default */
void LCD::Print(const char *psz, char line, bool fPad) const
{
	const char *pch, *pchLim;

	SetMemAddr(line == 1 ? 0x80 : 0xC0);
	pch = psz;
	while (*pch != 0)
		WriteCh(*pch++);
	if (fPad)
	{
		/* blank rest of line */
		pchLim = psz + cchScreen;
		while (pch++ < pchLim)
			WriteCh(' ');
	}
}

/*
DISPLAY DATA RAM ADDRESSES
CHARACTER 1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16
       + -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
LINE 1 | 80 81 82 83 84 85 86 87 88 89 8A 8B 8C 8D 8E 8F
LINE 2 | C0 C1 C2 C3 C4 C5 C6 C7 C8 C9 CA CB CC CD CE CF
*/
/* sets current "write" location to the specified position
   (ich, 0-based) on the specified line (1 or 2) */
void LCD::SetPos(char ich, char line) const
{
	SetMemAddr((line == 1 ? 0x80 : 0xC0) + ich);
}

/* writes a character at the current location, updates
   current location to next space (to right) */
void LCD::WriteCh(char ch) const
{
	WriteCommand(1, ch);
}

/* clears the display and resets the current location to the upper left */
void LCD::ClearAndReturn() const
{
	WriteCommand(0, 0x01);	// clear display
	WriteCommand(0, 0x02);	// return home (can take 1.52ms)
}

/* show the cursor (blinking or not) */
void LCD::ShowCursor(bool fBlink) const
{
	WriteCommand(0, fBlink ? 0x0F : 0x0E);
}

/* hide the cursor */
void LCD::HideCursor() const
{
	WriteCommand(0, 0x0C);
}
	
/* moves the cursor; fRight indicates direction, cch is
   count of characters to move [default = 1] */
void LCD::MoveCursor(bool fRight, char cch) const
{
	unsigned char grf;
	
	grf = fRight ? 0x14 : 0x10;
	while (cch-- > 0)
		WriteCommand(0, grf);
}

/* shifts the display; fRight indicates direction, cch is
   count of characters to shift [default = 1] */
void LCD::ShiftDisplay(bool fRight, char cch) const
{
	unsigned char grf;
	
	grf = fRight ? 0x1C : 0x18;
	while (cch-- > 0)
		WriteCommand(0, grf);
}

/*
*	helper functions
*/

void LCD::SetMemAddr(unsigned char ch) const
{
	WriteCommand(0, ch);
}


/* send an instruction to the LCD */
void LCD::WriteCommand(bool fRS, unsigned char ch) const
{
	/* make sure the processing for the previous request is done
	   (it can take up to 37+4us) */
	if (s_fInit)
	{
		/* set uC to read LCD's DB7 */
		DB7AsInput();

		/* request read from LCD; RS=0, R/!W=1 for read of BF (busy flag);
		   setting R/!W bit also disables (tri-states) output from 595 to LCD */
		RSForCmd();
		RWForRead();
		
		asm("nop");			// LCD wants wait (>=40ns) for data to be stable
		m_poutE->SetHigh();	// signal that we want to read
		/* wait for clear BF (busy flag) */
		while (DB7Val())
			;
		m_poutE->SetLow();		// done

		/* pulse again for 2nd nibble (we don't care about it, but we need to
		   read it because the LCD expects that!) */
		asm("nop");
		m_poutE->SetHigh();
		asm("nop");
		m_poutE->SetLow();
		
		/* set uC to write LCD's DB7 */
		DB7AsOutput();
	}
	else
	{
		/* can't check BF (busy flag), yet, so just to be safe,
		   delay for 1 ms here */
		g_timer.WaitMs(1);
	}

	WriteNibble(fRS, (ch >> 4) & 0xff);
	WriteNibble(fRS, ch & 0xff);
}

/* send a nibble */
void LCD::WriteNibble(bool fRS, unsigned char ch) const
{
	if (fRS)
		RSForData();
	else
		RSForCmd();
	RWForWrite();

	LCDDBAsOutput();

	PORTD &= ~mskPortD;
	PORTD |= (ch << shiftPortD) & mskPortD;
	PORTB &= ~mskPortB;
	PORTB |= (ch << shiftPortB) & mskPortB;

	/* all data is now waiting for LCD; needs to be stable
	   for >=40ns before E high */
	asm("nop");
	m_poutE->SetHigh();
	/* need to wait 230ns before E low
	   (nop = 1 cycle = 128ns at 8MHz)*/
	asm("nop");
	asm("nop");
	m_poutE->SetLow();
	
	/* make sure LCD inputs don't change for 10ns after E low */
	asm("nop");

	LCDDBAsInput();	// PB5/4/3 shared with pushbuttons
}
