/*	Motor.cpp
*
*	Motor control via H-bridge IC and PWM using Timer/Counter 1.
*
*	Written by Cathy Saxton
*	robotics@idleloop.com
*
*	Revisions:
*		10-26-05	disconnect OC1A/B output when speed=0
*/

#include <avr/io.h>
#include "main.h"

/*	A motor will turn when its inputs differ. The direction depends on which
	input is positive and which is ground. (If both are the same, the motor
	will stop moving!)

	When we go forward, we send a low (0) signal to the H-bridge for direction.
	We also send a PWM (pulse width modulation) signal to the H-bridge. When
	that signal is high, the motor turns; when it is low, the motor stops.

	So, we send a pulse that's high for the width specified, then drop it
	low for the remainder of the pulse frame, returning it to high at the
	end of that frame, when the next pulse starts.
	 ___   ___   ___
	|   |_|   |_|   |_		wave for 75% full speed forward

	When we go in reverse, we send a high (1) signal to the H-bridge for
	direction. In this case, the motor turns when the PWM signal is low,
	and stops when it's high.

	In this case, we sent a pulse that's low for the width specified, then
	raise it for the remainder of the frame.
	     _     _     _
	 ___| |___| |___| |		wave for 75% full speed in reverse


	The code below sets up Timer/Counter 1 as follows:
	- Waveform Generation Mode [WGM]: 8-bit Fast PWM (0101) -- counting
			from 0-255 (then wrapping)
	- Compare Output Mode [COM]: this determines how the signals on OC1A/PB1
				(right motor) and OC1B/PB2 (left motor) change when the
				counter matches the value in OCR1A or OCR1B (our "PWM"
				values indicating the desired speeds of the motors)
				and at the maximum count value (255, end of frame).
			clear on match, set at TOP (10) [for forward, see above], or
			set on match, clear at TOP (11) [for reverse, see above]
	- Clock Select: clkIO/8 (010)
			8 Mhz / 8 (prescaler) = 1 MHz for counter
			1 MHz / 256 (clock pulses/frame) = 4 kHz for PWM [frame ~256 us]

	See the Atmel ATmega8 data sheet for details on these settings and how
	they correspond to bits in TCCR1A/B (Timer/Counter 1 Control Registers).
*/

/* defines to help set COM bits in TCCR1A */
#define mskCom			3	// (11) mask for COM bits
#define comFwd			2	// (10) COM bits for forward motion
#define comRev			3	// (11)	COM bits for reverse direction
#define shComA			6	// bits to shift mskCom for COM1A bits (**00 0000)
#define shComB			4	// bits to shift mskCom for COM1B bits (00** 0000)

bool MOTOR::s_fInit = false;

MOTOR::MOTOR(OUT *poutPWM, OUT *poutDir)
	: m_poutDir(poutDir)
{
	if (poutPWM->m_bit == 0x01 << 2)		// B2 = OC1B
	{
		m_pregOCR1xL = &OCR1BL;
		m_shCom = shComB;
	}
	else
	{
		m_pregOCR1xL = &OCR1AL;
		m_shCom = shComA;
	}
	
	if (!s_fInit)
		Init();

	poutPWM->SetLow();
	poutDir->SetLow();
}

/* sets up microcontroller for PWM control of motors */
void MOTOR::Init()
{
	s_fInit = true;

	TCNT1 = 0;

	/* see comment above for info on PWM initialization */
	/* start with motors disconnected from Timer/Counter output */
	TCCR1A = 0x01;	// 00 00 00 01
	TCCR1B = 0x0A;	// 000 01 010
	
	/* OCR1A/B are the values that the timer is compared to; a match will
	   cause the output to change; small values mean the motor runs for a
	   short period (slower); larger values are longer times (faster)*/
	OCR1A = OCR1B = 0;	// (value is irrelevant since outputs are disconnected)
}

void MOTOR::SetSpeed(int pwm) const
{
	TCCR1A &= ~(mskCom << m_shCom);		// clear COM bits
	if (pwm == 0)
	{
		/* stop -- disconnect output on OC1A/B; it should go
		   back to being an output with value of 0 */
		/* COM bits are cleared, which is what we need for disconnect */
		m_poutDir->SetLow();
	}
	else
	{
		if (pwm > 0)
		{
			/* forward */
			TCCR1A |= comFwd << m_shCom;
			m_poutDir->SetLow();
		}
		else
		{
			/* reverse */
			TCCR1A |= comRev << m_shCom;
			m_poutDir->SetHigh();
	
			pwm = -pwm;
		}
		if (pwm > 255)
			pwm = 255;
		
		*m_pregOCR1xL = (char)pwm;
	}
}
