/*	Timer.cpp
*
*	Accurate timing.
*
*	NOTE: this code...
*	- uses Timer/Counter 2
*	- assumes internal clock is running at 8MHz
*
*	Written by Cathy Saxton
*	robotics@idleloop.com
*
*	Revisions:
*		6-29-06	signal.h/INTERRUPT() -> interrupt.h/ISR() [new avr-gcc]
*				change timing (k = 1000, not 1024 for Hz)
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include "main.h"

bool TIMER::s_fInit = false;

/* this counts the number of milliseconds since Timer/Counter 2
   was initialized (volatile to make sure loops work!) */
static volatile unsigned long s_msTimer = 0;

TIMER::TIMER()
{
	if (!s_fInit)
		Init();
}

/* sets up Timer/Counter 2 (assumes internal clock at 8MHz) */
void TIMER::Init()
{
	s_fInit = true;

	/* start counting at 0 */
	TCNT2 = 0;

	/* CS (Clock Select) /64 [100] -- 8 Mhz / 64 = 125kHz 
	   CTC (Clear Timer on Compare Match) waveform generation mode [10]
	   OCR2 (top/clear value) = 125
	   for match every 1 ms
	*/
	TCCR2 = 0x0c;	// 0 0 00 1 100
	OCR2 = 125;

	/* set Compare Match Interrupt Enable (for timer/counter 2) */
	SetBit(TIMSK, OCIE2);

	/* set I-bit in SREG (global enable for interrupts) */
	sei();
}

/* returns number of ms since Timer initialized */
unsigned long TIMER::MsCur() const
{
	char sregSav;
	unsigned long ms;

	/* disable (delay) interrupts while we read
	   s_msTimer since it's 4 bytes and we don't
	   want it changing during the read! */
	sregSav = SREG;	// cache register holding Global Interrupt Flag
	cli();			// clear Global Interrupt Flag
	ms = s_msTimer;
	SREG = sregSav;	// restore register holding Global Interrupt Flag

	return ms;
}

/* returns true if the specified number of milliseconds
   has passed since the start time */
bool TIMER::HasElapsed(unsigned long msStart, unsigned int msWait) const
{
	return MsCur() - msStart > msWait;
}

/* waits (pauses) for the specified number of milliseconds */
void TIMER::WaitMs(unsigned int ms) const
{
	unsigned long msStart;

	msStart = MsCur();
	while (!HasElapsed(msStart, ms))
		;
}

/* this interrupt happens every ms, when TCNT2 reaches OCR2 */
ISR(TIMER2_COMP_vect)
{
	++s_msTimer;
}
