/*	main.cpp
*
*	Sample program for Pololu's Orangutan Robot Controller.
*
*	Written by Cathy Saxton
*	robotics@idleloop.com
*/

#include <stdio.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include "main.h"

/*
*	Globals
*
*	These lines declare variables for using inputs, outputs, and
*	microcontroller features. Initialization of the appropriate
*	ports/pins is done by the class constructors. Declaring them
*	outside main() makes them globals, accessible from all
*	functions.
*/

char g_szBuf[17];

TIMER g_timer;

OUT    g_outBuzz ('B', 0, 0);       // buzzer
OUT    g_mtrPWMR ('B', 1, 0);       // right motor PWM (speed) [OC1A]
OUT    g_mtrPWML ('B', 2, 0);       // left motor PWM (speed) [OC1B]
IN     g_btnL    ('B', 3, false);	// also LCD DB4
IN     g_btnR    ('B', 4, false);	// also LCD DB5
IN     g_btnSel  ('B', 5, false);	// also LCD DB6
/* (C6 is RESET) */
OUT    g_led     ('D', 1, 0);	// red LED on Baby Orangutan board
OUT    g_lcdRS   ('D', 2, 0);	// LCD: RS
OUT    g_lcdRW   ('D', 3, 0);	// LCD: RW
OUT    g_lcdE    ('D', 4, 0);	// LCD: E
OUT    g_mtrDirR ('D', 5, 0);	// right motor direction (forward/reverse)
OUT    g_mtrDirL ('D', 6, 0);	// left motor direction
/* D7 is LCD DB7 */

ANALOG g_adcPot(7);		// ADC7 is connected to potentiometer for variable V input
BUZZER g_buzzer(&g_outBuzz);
LCD    g_lcd(&g_lcdRS, &g_lcdRW, &g_lcdE);
MOTOR  g_motorL(&g_mtrPWML, &g_mtrDirL);
MOTOR  g_motorR(&g_mtrPWMR, &g_mtrDirR);

/*
*	Definitions
*/
#define addrBeepPref	0		// eeprom address where beep preference is stored
#define chBeepOn		'B'		// character stored when button beep is turned on
#define chBeepOff		0		// character stored when button beep is turned off

/*
*	Menu
*
*	The menu data is stored in "program memory" (Flash). The default
*	location without the "PROGMEM" specification would be in SRAM,
*	taking up precious run-time memory space. To use data stored in
*	Flash memory, it is first copied to a location in SRAM, only
*	requiring enough space in SRAM for a single menu item, not the
*	whole list.
*
*	To add an item to the menu, list its prototype (with the others)
*	and add an entry to the s_ami array, using the existing items
*	as an example. 
*/

typedef void (*PFN_MENU_ITEM)(void);

struct MI	// Menu Item
{
	char sz[9];
	PFN_MENU_ITEM pfn;
};

void DoBlink();
void DoBlink2();
void DoBuzzer();
void DoDrive();
void DoTimer();
void SetButtonBeep();

MI s_ami[] PROGMEM =
{
	/* 12345678 -- max length! */
	{ "Blink", DoBlink },
	{ "Blink2", DoBlink2 },
	{ "Buzzer", DoBuzzer },
	{ "Drive", DoDrive },
	{ "Timer", DoTimer },
	{ "Set Beep", SetButtonBeep },
};
#define cmi (sizeof(s_ami) / sizeof(s_ami[0]))

char s_szUsage[] PROGMEM = "< > sel";

/*
*	main program
*/

int main()
{
	int imi, imiPrev;
	MI mi;
	
	imi = 0;
	imiPrev = -1;
	
	/* see if user has set a preference for beeps on button press;
	   it's stored in eeprom */
	if (EEPROMReadByte(addrBeepPref) == chBeepOn)
	{
		g_btnL.SetBeep(true);
		g_btnR.SetBeep(true);
		g_btnSel.SetBeep(true);
	}
	
	while (true)
	{
		if (imiPrev != imi)
		{
			/* show current selection and usage info */
			memcpy_P(&mi, &s_ami[imi], sizeof(MI));
			g_lcd.Print(mi.sz, 1);
			memcpy_P(g_szBuf, s_szUsage, sizeof(s_szUsage));
			g_lcd.Print(g_szBuf, 2);
			imiPrev = imi;
		}
		if (g_btnSel.WasPressed())
		{
			g_lcd.Print("", 1);
			g_lcd.Print("", 2);
			g_timer.WaitMs(500);	// pause to give time to remove finger from button!
			(mi.pfn)();
			imiPrev = -1;
		}
		else if (g_btnL.WasPressed())
		{
			if (--imi < 0)
				imi = cmi-1;
		}
		else if (g_btnR.WasPressed())
		{
			if (++imi >= (int)cmi)
				imi = 0;
		}
	}

	return 0;
}

/*
*	LED examples
*/

/* simple blink, once a second; hold sel for at least a second to exit */
void DoBlink()
{
	/* note: it would be more efficient usage of SRAM (run-time memory)
	   to use the "PROGMEM" stuff like above for Usage, but this example
	   shows a simpler way to do LCD output */
	g_lcd.Print("push sel", 1);
	g_lcd.Print("to exit", 2);
	while (!g_btnSel.WasPressed())
	{
		g_led.On();
		g_timer.WaitMs(200);
		g_led.Off();
		g_timer.WaitMs(800);
	}
}

/* blink at a rate based on the input from the potentiometer hooked to ADC7;
   press sel to exit */
void DoBlink2()
{
	uchar adc;
	ulong msStart;
	uint msDelay;
	
	g_lcd.Print("push sel", 2);
	g_led.On();
	msStart = g_timer.MsCur();
	while (!g_btnSel.WasPressed())
	{
		adc = g_adcPot.ReadByte();
		msDelay = adc * 4;
		
		/* note: this is pretty cool, but "snprintf" uses almost 2k of flash! */
		snprintf(g_szBuf, sizeof(g_szBuf), "pot: %3d", adc);
		g_lcd.Print(g_szBuf, 1);
		if (g_timer.HasElapsed(msStart, msDelay))
		{
			g_led.Toggle();
			msStart = g_timer.MsCur();
		}
	}
	g_led.Off();
}

/*
*	Buzzer example
*/

/* sound buzzer; press button to select frequency (1 is highest pitch) */
void DoBuzzer()
{
	uchar tone;
	
	g_lcd.Print("Freq...", 1);
	g_lcd.Print("1  2  3", 2);
	
	tone = toneMax;
	do
	{
		if (g_btnL.WasPressed())
			tone = toneHigh;
		if (g_btnR.WasPressed())
			tone = toneMid;
		if (g_btnSel.WasPressed())
			tone = toneLow;
	}
	while (tone == toneMax);

	g_buzzer.Buzz(tone, 300);
}

/*
*	Motor / driving example
*/

/* drive forward, then back (then stop!) */
void DoDrive()
{
	g_motorL.SetSpeed(128);
	g_motorR.SetSpeed(128);
	g_timer.WaitMs(1500);

	g_led.On();
	g_motorL.SetSpeed(-128);
	g_motorR.SetSpeed(-128);
	g_timer.WaitMs(1500);
	g_led.Off();

	g_motorL.SetSpeed(0);
	g_motorR.SetSpeed(0);
}

/*
*	Timer example
*/
void DoTimer()
{
	ulong msStart, msT;
	ushort ms;
	uchar sec, min;

	g_lcd.Print("sel=exit", 2);

	msStart = g_timer.MsCur();
	do
	{
		msT = g_timer.MsCur() - msStart;
		ms = (ushort)(msT % 1000);
		msT /= 1000;
		sec = (uchar) msT % 60;
		min = (uchar) msT / 60;
		/* m:ss.xxx (2-digit minutes will just shift display and lose 1000's place) */
		snprintf(g_szBuf, sizeof(g_szBuf), "%d:%02d.%03d", min, sec, ms);
		g_lcd.Print(g_szBuf, 1);
	}
	while (!g_btnSel.WasPressed());
}

/*
*	EEPROM example
*
*	also see code at the beginning of main()
*/

void SetButtonBeep()
{
	char btn;
	
	g_lcd.Print("btn beep", 1);
	g_lcd.Print("on   off", 2);

	btn = 0;
	do
	{
		if (g_btnL.WasPressed())
			btn = 1;
		if (g_btnSel.WasPressed())
			btn = 2;
	}
	while (btn == 0);

	if (btn == 1)	// on
	{
		g_btnL.SetBeep(true);
		g_btnR.SetBeep(true);
		g_btnSel.SetBeep(true);
		EEPROMWriteByte(addrBeepPref, chBeepOn);
	}
	else
	{
		g_btnL.SetBeep(false);
		g_btnR.SetBeep(false);
		g_btnSel.SetBeep(false);
		EEPROMWriteByte(addrBeepPref, chBeepOff);
	}
}
